from .routes import instance_config_routes, role_routes, datacenter_routes, publish_routes, application_routes, grant_routes
import aiohttp
import datetime as dt
import asyncio

import Library.utils.convert as cd

from ..utils.DictDot import DictDot
from dataclasses import dataclass, field
from typing import List

import importlib

from .DomoAuth import DomoFullAuth
from .DomoGrant import DomoGrant
from .DomoRole import DomoRole
from .DomoApplication import DomoApplication

import Library.DomoClasses.DomoPublish as dmpb
importlib.reload(dmpb)


@dataclass
class DomoConnector:
    id: str
    label: str
    title: str
    sub_title: str
    description: str
    create_date: dt.datetime
    last_modified: dt.datetime
    publisher_name: str
    writeback_enabled: bool
    tags: list[str] = field(default_factory=list)
    capabilities: list[str] = field(default_factory=list)

    @classmethod
    def _from_str(cls, obj):
        dd = DictDot(obj)

        return cls(
            id=dd.databaseId,
            label=dd.label,
            title=dd.title,
            sub_title=dd.subTitle,
            description=dd.description,
            create_date=cd.convert_epoch_millisecond_to_datetime(
                dd.createDate),
            last_modified=cd.convert_epoch_millisecond_to_datetime(
                dd.lastModified),
            publisher_name=dd.publisherName,
            writeback_enabled=dd.writebackEnabled,
            tags=dd.tags,
            capabilities=dd.capabilities
        )


@dataclass
class DomoInstanceConfig:
    full_auth: DomoFullAuth

    @classmethod
    def _authorized_domain_from_string(cls, authorized_domain_string):
        if authorized_domain_string == "":
            return []

        # return json.loads(authorized_domain_string)

        return authorized_domain_string.split(',')

    @classmethod
    async def get_whitelist(cls, full_auth: DomoFullAuth, session: aiohttp.ClientSession = None, debug: bool = False) -> List[str]:
        res = await instance_config_routes.get_whitelist(full_auth=full_auth, session=session, debug=debug)

        if res.status == 200:
            return res.response.get('addresses')

    @classmethod
    async def update_whitelist(cls, full_auth: DomoFullAuth, ip_address_list: list[str], debug: bool = False, is_upsert: bool = True, log_results: bool = False):

        res = await instance_config_routes.update_whitelist(full_auth=full_auth,
                                                            ip_address_list=ip_address_list,
                                                            debug=debug, log_results=log_results)

        if debug:
            print(res)

        if res.status == 200:
            return {'status': res.status, 'response': res.response}

    @classmethod
    async def get_roles(cls, full_auth: DomoFullAuth,
                        debug: bool = False,
                        session: aiohttp.ClientSession = None):

        res = await role_routes.get_roles(full_auth=full_auth,
                                          debug=debug,
                                          session=session)

        if res.status == 200:
            json_list = res.response
            return [DomoRole._from_str(id=obj.get('id'),
                                       name=obj.get('name'),
                                       description=obj.get('description'),
                                       full_auth=full_auth
                                       ) for obj in json_list]

    @classmethod
    async def get_grants(cls, full_auth: DomoFullAuth,
                         debug: bool = False,
                         session: aiohttp.ClientSession = None,
                         return_raw: bool = False):

        res = await grant_routes.get_grants(full_auth=full_auth,
                                            debug=debug,
                                            session=session)

        if res.status == 200 and not return_raw:
            json_list = res.response
            return [DomoGrant._from_json(obj) for obj in json_list]

        elif res.status == 200 and return_raw:
            return res.response

    @classmethod
    async def get_authorized_domains(cls, full_auth: DomoFullAuth, debug: bool = False, session: aiohttp.ClientSession = None):
        res = await instance_config_routes.get_authorized_domains(full_auth=full_auth,
                                                                  debug=debug,
                                                                  session=session)

        if res.status == 200:
            str = cls._authorized_domain_from_string(res.response.get('value'))
            return str

    @classmethod
    async def get_connectors(cls, full_auth: DomoFullAuth,
                             session: aiohttp.ClientSession = None,
                             debug: bool = False,
                             limit=100,
                             ):

        is_close_session = False
        if not session:
            is_close_session = True
            session = aiohttp.ClientSession()

        def arr_fn(res):
            # pprint(res.response)
            return res.response.get('searchObjects')

        def alter_maximum_fn(res):
            return res.response.get('totalResultCount')

        body = {
            "count": limit,
            "offset": 0,
            "hideSearchObjects": True,
            "combineResults": False,
            "entities": ["CONNECTOR"],
            "query": "*"}

        obj_list = await datacenter_routes.search_datacenter(
            full_auth=full_auth,
            arr_fn=arr_fn,
            alter_maximum_fn=alter_maximum_fn,
            body=body,
            session=session,
            limit=limit,
            debug=debug)

        if is_close_session:
            await session.close()

        return [DomoConnector._from_str(obj) for obj in obj_list]

    @classmethod
    async def update_authorized_domains(cls, full_auth: DomoFullAuth,
                                        authorized_domain_list: list[str],
                                        is_replace_existing_list: bool = False,
                                        debug: bool = False):

        if not is_replace_existing_list:
            existing_domain_list = await cls.get_authorized_domains(full_auth=full_auth, debug=debug)

            authorized_domain_list.extend(existing_domain_list)

        if debug:
            print(
                f'🌡️ updating authorized domain with {",".join(authorized_domain_list)}')

        res = await instance_config_routes.update_authorized_domains(full_auth=full_auth,
                                                                     authorized_domain_list=authorized_domain_list,
                                                                     debug=debug)

        if debug:
            "update_authorized_domains"
            print(res)

        if res.status == 200 or res.status == 204:
            res = {'authorized_domains': await cls.get_authorized_domains(full_auth=full_auth, debug=debug),
                   'status': 200}

        return res

    @classmethod
    async def get_applications(cls,
                               full_auth: DomoFullAuth,
                               debug: bool = False, session: aiohttp.ClientSession = None, return_raw: bool = False):

        res = await application_routes.get_applications(full_auth=full_auth,
                                                        debug=debug,
                                                        session=session)
        if debug:
            print('Getting Domostats jobs')

        if res.status == 200 and not return_raw:
            return [DomoApplication._from_json(job) for job in res.response]

        if res.status == 200 and return_raw:
            return res.response

    @classmethod
    async def get_publications(cls,
                               full_auth: DomoFullAuth,
                               debug: bool = False, session: aiohttp.ClientSession = None, return_raw: bool = False):

        res = await publish_routes.search_publications(full_auth=full_auth,
                                                       debug=debug,
                                                       session=session)
        if debug:
            print('Getting Publish jobs')

        if res.status == 200 and not return_raw:
            return await asyncio.gather(*[dmpb.DomoPublication.get_from_id(publication_id=job.get('id'),
                                                                           full_auth=full_auth) for job in res.response])

        if res.status == 200 and return_raw:
            return res.response



import aiohttp
import asyncio
from enum import Enum

from dataclasses import dataclass, field

import Library.DomoClasses.DomoAuth as dmda
import Library.DomoClasses.DomoDatacenter as dmdc

from .routes import datacenter_routes


def get_content_list_ls(cls_obj, regex_pattern_ls=None):
    import re

    regex_pattern_ls = regex_pattern_ls or ['.*_id_ls$', '^content_.*']

    content_list_ls = [content_list for content_list in dir(cls_obj)
                       if all([re.match(pattern, content_list) for pattern in regex_pattern_ls])]

    result = []
    for content_name in content_list_ls:
        if not getattr(cls_obj, content_name) or len(getattr(cls_obj, content_name)) == 0:
            continue

        base_name = content_name
        [base_name := re.sub(regex_pattern.replace('.*', ''), '', base_name)
         for regex_pattern in regex_pattern_ls]

        result.append({'list_name': content_name,
                       'entity_name': base_name,
                       'regex_pattern_ls': regex_pattern_ls
                       })
    return result


@dataclass
class DomoLineage:
    id: str
    parent: any = field(default=None, repr=False)
    page_id_ls: [str] = None
    card_id_ls: [str] = None
    dataflow_id_ls: [str] = None
    dataset_id_ls: [str] = None

    entity_ls: [any] = None

    async def _get_page_card_ids(self):

        import Library.DomoClasses.DomoPage as dmpg

        if not self.parent.content_page_id_ls or len(self.parent.content_page_id_ls) == 0:
            return None

        page_card_ls = await asyncio.gather(*[dmpg.DomoPage.get_cards(page_id=page_id,
                                                                      full_auth=self.parent.full_auth)
                                              for page_id in self.parent.content_page_id_ls])

        if not page_card_ls or len(page_card_ls) == 0:
            return

        if not self.card_id_ls:
            self.card_id_ls = []

        for page in page_card_ls:
            if page and len(page) > 0:
                for card in page:
                    if card.id not in self.card_id_ls:
                        self.card_id_ls.append(card.id)

        return self.card_id_ls

    async def get_entity_lineage_upstream(self,
                                          entity_id,
                                          entity_type,
                                          full_auth: dmda.DomoFullAuth = None,
                                          session: aiohttp.ClientSession = None,
                                          debug: bool = False,
                                          debug_prn: bool = False):

        import Library.DomoClasses.DomoDataflow as dmdf
        import Library.DomoClasses.DomoDataset as dmds

        try:
            full_auth = full_auth or self.parent.full_auth

            if not session:
                is_close_session = True
            session = session or aiohttp.ClientSession()

            res = await datacenter_routes.get_lineage_upstream(full_auth=full_auth,
                                                               entity_type=entity_type,
                                                               entity_id=entity_id,
                                                               session=session, debug=debug)
            if res.status != 200:
                return None

            obj = res.response

            domo_obj_ls = []

            for key, item in obj.items():
                entity_type = item.get('type')
                entity_id = item .get('id')

                if not self.entity_ls:
                    self.entity_ls = []

                if entity_type == 'DATA_SOURCE':
                    if not self.dataset_id_ls:
                        self.dataset_id_ls = []

                    if entity_id not in self.dataset_id_ls:
                        self.dataset_id_ls.append(entity_id)

                        do = await dmds.DomoDataset.get_from_id(full_auth=full_auth, id=entity_id)
                        domo_obj_ls.append(do)
                        self.entity_ls.append(do)

                if entity_type == 'DATAFLOW':
                    if not self.dataflow_id_ls:
                        self.dataflow_id_ls = []

                    if entity_id not in self.dataflow_id_ls:
                        self.dataflow_id_ls.append(entity_id)

                        do = await dmdf.DomoDataflow.get_from_id(full_auth=full_auth, id=entity_id)
                        domo_obj_ls.append(do)
                        self.entity_ls.append(do)

            return domo_obj_ls

        finally:
            if is_close_session:
                await session.close()

    async def _get_entity_ls_lineage(self,
                                     domo_entity: dmdc.DomoEntity,
                                     full_auth=None,
                                     session: dmda.DomoFullAuth = None,
                                     debug: bool = False, debug_prn: bool = False):
        full_auth = full_auth or self.full_auth

        entity_attribute = f"{domo_entity.name.lower()}_id_ls"

        return await asyncio.gather(*[self.get_entity_lineage_upstream(full_auth=full_auth,
                                                                       entity_id=entity_id,
                                                                       entity_type=domo_entity.value,
                                                                       session=session,
                                                                       debug=debug, debug_prn=debug_prn)
                                      for entity_id in getattr(self, entity_attribute)])

    def _reset_lineage_and_sync_parent(self):

        content_list = get_content_list_ls(self.parent)

        for content_obj in content_list:

            parent_content = getattr(self.parent, content_obj.get('list_name'))

            lineage_content_name = f"{content_obj.get('entity_name')}_id_ls"

            setattr(self,
                    lineage_content_name,
                    parent_content
                    )
        return self

    async def get(self,
                  debug_prn: bool = False, debug: bool = False,
                  full_auth: dmda.DomoFullAuth = None,
                  session=None):

        full_auth = full_auth or self.parent.full_auth

        self._reset_lineage_and_sync_parent()

        if self.page_id_ls:
            await self._get_page_card_ids()

        if self.card_id_ls and len(self.card_id_ls) > 0:
            if debug_prn:
                print(f'🏁 getting card lineage for repo {self.id}')
            await self._get_entity_ls_lineage(domo_entity=dmdc.DomoEntity.CARD,
                                              full_auth=full_auth,
                                              debug_prn=debug_prn, debug=debug, session=session)

        if self.dataflow_id_ls and len(self.dataflow_id_ls) > 0:
            if debug_prn:
                print(f'🏁 getting dataflow lineage for repo {self.id}')

            await self._get_entity_ls_lineage(domo_entity=dmdc.DomoEntity.DATAFLOW,
                                              full_auth=full_auth,
                                              debug_prn=debug_prn, debug=debug, session=session)

        if self.dataset_id_ls and len(self.dataset_id_ls) > 0:
            if debug_prn:
                print(f'🏁 getting dataset lineage for repo {self.id}')

            await self._get_entity_ls_lineage(domo_entity=dmdc.DomoEntity.DATASET,
                                              full_auth=full_auth,
                                              debug_prn=debug_prn, debug=debug, session=session)

        return self

    def _flatten_lineage(self):
        attribute_ls = get_content_list_ls(self, ['.*_id_ls$'])

        output_ls = []

        for attribute in attribute_ls:
            ls_name = attribute.get('list_name')
            entity_name = attribute.get('entity_name')
            entity_type = dmdc.DomoEntity[entity_name.upper()].value

            row_ls = [{'entity_type': entity_type,
                       'entity_id': row} for row in getattr(self, ls_name)]
            output_ls += row_ls

        return output_ls



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/integrations/MonitDataset.ipynb.

# %% auto 0
__all__ = []

# %% ../../nbs/integrations/MonitDataset.ipynb 2
from dataclasses import dataclass, field
import pandas as pd
import datetime as dt
import decimal
import domolibrary.classes.DomoDataset as dcds
import domolibrary.classes.DomoStream as dcdstr


@dataclass
class MonitDataset_Record:
    id: str = None
    stream_account_id: int = None
    stream_account_userid: int = None
    stream_account_display_name: str = None
    nb_partitions: int = 0
    name: str = None
    data_provider_type: str = None
    display_type: str = None
    stream_id: str = None
    stream_config_tables: str = None
    stream_config_query: str = None
    instance: str = None
    owner: str = None
    row_count: int = None
    url: str = None
    report_date: dt.datetime = dt.datetime.now()
    tags: list = field(default=None)
    is_certified: int = 0
    certification_name: str = None
    connector_version: decimal.Decimal(2) = None
    update_method: str = None
    connector_id: str = None


@dataclass
class MonitDataset_Data:
    ds: dcds.DomoDataset = None
    stream_ds: dcdstr.DomoStream = None
    partition_ls: list = None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccount.ipynb.

# %% auto 0
__all__ = ['DomoAccount_Config', 'DomoAccount_Config_AbstractCredential', 'DomoAccount_Config_DatasetCopy',
           'DomoAccount_Config_Governance', 'DomoAccount_Config_AmazonS3', 'DomoAccount_Config_AmazonS3Advanced',
           'DomoAccount_Config_AwsAthena', 'DomoAccount_Config_HighBandwidthConnector', 'AccountConfig', 'DomoAccount',
           'DomoAccount_DataProviderType_ConfigNotDefined', 'DomoAccount_UpdateName_Error',
           'DomoAccount_CreateAccount_Error', 'DomoAccount_DeleteAccount_Error', 'DomoAccounts']

# %% ../../nbs/classes/50_DomoAccount.ipynb 3
from ..routes.account import ShareAccount_V1_AccessLevel, ShareAccount_V2_AccessLevel, ShareAccount


# %% ../../nbs/classes/50_DomoAccount.ipynb 4
import asyncio

from enum import Enum
from dataclasses import dataclass, field
from abc import ABC, abstractmethod

from typing import Union

import datetime as dt
import re


import httpx

from fastcore.basics import patch_to

import domolibrary.utils.convert as cd
import domolibrary.utils.DictDot as util_dd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de
import domolibrary.routes.account as account_routes


# %% ../../nbs/classes/50_DomoAccount.ipynb 6
class DomoAccount_Config(ABC):
    """DomoAccount Config abstract base class"""

    data_provider_type: str

    @classmethod
    @abstractmethod
    def _from_json(cls, obj):
        """convert accounts API response into a class object"""
        pass

    @abstractmethod
    def to_json(self):
        """convert class object into a format the accounts API expects"""
        pass


# %% ../../nbs/classes/50_DomoAccount.ipynb 8
@dataclass
class DomoAccount_Config_AbstractCredential(DomoAccount_Config):
    data_provider_type = "abstract-credential-store"
    credentials: dict

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            credentials=dd.credentials,
        )

    def to_json(self):
        return {"credentials": self.credentials}


# %% ../../nbs/classes/50_DomoAccount.ipynb 9
@dataclass
class DomoAccount_Config_DatasetCopy(DomoAccount_Config):
    domo_instance: str
    access_token: str = field(repr=False)

    data_provider_type = "dataset-copy"

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(access_token=dd.accessToken, domo_instance=dd.instance)

    def to_json(self):
        return {"accessToken": self.access_token, "instance": self.domo_instance}


# %% ../../nbs/classes/50_DomoAccount.ipynb 10
@dataclass
class DomoAccount_Config_Governance(DomoAccount_Config):
    domo_instance: str
    access_token: str = field(repr=False)

    data_provider_type = "domo-governance-d14c2fef-49a8-4898-8ddd-f64998005600"

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(access_token=dd.apikey, domo_instance=dd.customer)

    def to_json(self):
        return {"apikey": self.access_token, "customer": self.domo_instance}


# %% ../../nbs/classes/50_DomoAccount.ipynb 12
@dataclass
class DomoAccount_Config_AmazonS3(DomoAccount_Config):
    access_key: str
    secret_key: str = field(repr=False)
    bucket: str
    region: str = "us-west-2"
    data_provider_type = "amazon-s3"

    
    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            access_key=dd.accessKey,
            secret_key=dd.secretKey,
            bucket=dd.bucket,
            region=dd.region,
        )

    def to_json(self):
        if self.bucket.lower().startswith("s3://"):
            bucket = self.bucket[5:]
            print(f"🤦‍♀️- Domo bucket expects string without s3:// prefix. Trimming to '{bucket}' for the output")
        return {
            "accessKey": self.access_key,
            "secretKey": self.secret_key,
            "bucket": bucket,
            "region": self.region,
        }

# %% ../../nbs/classes/50_DomoAccount.ipynb 13
@dataclass
class DomoAccount_Config_AmazonS3Advanced(DomoAccount_Config):
    access_key: str
    secret_key: str = field(repr=False)
    bucket: str
    region: str = "us-west-2"
    data_provider_type = "amazons3-advanced"

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            access_key=dd.accessKey,
            secret_key=dd.secretKey,
            bucket=dd.bucket,
            region=dd.region,
        )

    def to_json(self):
        if self.bucket.lower().startswith("s3://"):
            bucket = self.bucket[5:]
            print(
                f"🤦‍♀️- Domo bucket expects string without s3:// prefix. Trimming to '{bucket}' for the output")
        return {
            "accessKey": self.access_key,
            "secretKey": self.secret_key,
            "bucket": bucket,
            "region": self.region,
        }


# %% ../../nbs/classes/50_DomoAccount.ipynb 14
@dataclass
class DomoAccount_Config_AwsAthena(DomoAccount_Config):
    aws_access_key: str
    aws_secret_key: str = field(repr=False)
    s3_staging_dir: str
    workgroup: str

    region: str = "us-west-2"
    data_provider_type = "aws-athena"

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            aws_access_key=dd.awsAccessKey,
            aws_secret_key=dd.awsSecretKey,
            s3_staging_dir=dd.s3StagingDir,
            region=dd.region,
            workgroup=dd.workgroup,
        )

    def to_json(self):
        return {
            "awsAccessKey": self.aws_access_key,
            "awsSecretKey": self.aws_secret_key,
            "s3StagingDir": self.s3_staging_dir,
            "region": self.region,
            "workgroup": self.workgroup,
        }

# %% ../../nbs/classes/50_DomoAccount.ipynb 15
@dataclass
class DomoAccount_Config_HighBandwidthConnector(DomoAccount_Config):
    """ this connector is not enabled by default contact your CSM / AE"""
    
    aws_access_key: str
    aws_secret_key: str = field(repr=False)
    s3_staging_dir: str

    region: str = "us-west-2"
    data_provider_type = "amazon-athena-high-bandwidth"

    @classmethod
    def _from_json(cls, obj):

        dd = util_dd.DictDot(obj)

        return cls(
            aws_access_key=dd.awsAccessKey,
            aws_secret_key=dd.awsSecretKey,
            s3_staging_dir=dd.s3StagingDir,
            region=dd.region,
        )

    def to_json(self):
        return {
            "awsAccessKey": self.aws_access_key,
            "awsSecretKey": self.aws_secret_key,
            "s3StagingDir": self.s3_staging_dir,
            "region": self.region,
        }


# %% ../../nbs/classes/50_DomoAccount.ipynb 18
class AccountConfig(Enum):
    """
    Enum provides appropriate spelling for data_provider_type and config object.
    The name of the enum should correspond with the data_provider_type with hyphens replaced with underscores.
    """
    abstract_credential_store = DomoAccount_Config_AbstractCredential

    dataset_copy = DomoAccount_Config_DatasetCopy

    domo_governance_d14c2fef_49a8_4898_8ddd_f64998005600 = DomoAccount_Config_Governance

    aws_athena = DomoAccount_Config_AwsAthena
    amazon_athena_high_bandwidth = DomoAccount_Config_HighBandwidthConnector

    amazon_s3 = DomoAccount_Config_AmazonS3
    amazons3_advanced = DomoAccount_Config_AmazonS3Advanced


# %% ../../nbs/classes/50_DomoAccount.ipynb 20
@dataclass
class DomoAccount:
    name: str
    data_provider_type: str

    id: int = None
    created_dt: dt.datetime = None
    modified_dt: dt.datetime = None
    auth: dmda.DomoAuth = field(repr=False, default=None)

    config: DomoAccount_Config = None

    @classmethod
    def _from_json(cls, obj: dict, auth: dmda.DomoAuth = None):
        """converts data_v1_accounts API response into an accounts class object"""

        dd = util_dd.DictDot(obj)

        return cls(
            id=dd.id or dd.databaseId,
            name=dd.displayName,
            data_provider_type=dd.dataProviderId or dd.dataProviderType,
            created_dt=cd.convert_epoch_millisecond_to_datetime(dd.createdAt or dd.createDate),
            modified_dt=cd.convert_epoch_millisecond_to_datetime(dd.modifiedAt or dd.lastModified),
            auth=auth,
        )

# %% ../../nbs/classes/50_DomoAccount.ipynb 21
class DomoAccount_DataProviderType_ConfigNotDefined(de.DomoError):
    def __init__(
        self, account_id, data_provider_type, domo_instance, function_name="_get_config"
    ):

        message = f"🛑 data provider type {data_provider_type} for account_id {account_id} not defined yet.  Extend the AccountConfig class"

        super().__init__(
            message=message, function_name=function_name, domo_instance=domo_instance
        )


@patch_to(DomoAccount)
async def _get_config(
    self: DomoAccount, session=None, 
    return_raw: bool = False, 
    debug_api: bool = None, 
    debug_prn :bool = False
):

    res_config = await account_routes.get_account_config(
        auth=self.auth,
        account_id=self.id,
        data_provider_type=self.data_provider_type,
        session=session,
        debug_api=debug_api,
    )

    if return_raw:
        return res_config

    enum_clean = re.sub("-", "_", self.data_provider_type)

    if debug_prn:
        print(f'retrieving config for {self.id} - {self.name} with {res_config.response}')

    if not enum_clean in AccountConfig.__members__:
        raise DomoAccount_DataProviderType_ConfigNotDefined(
            account_id=self.id,
            data_provider_type=self.data_provider_type,
            domo_instance=self.auth.domo_instance,
            function_name = '_get_config'
        )

    self.config = AccountConfig[enum_clean].value._from_json(res_config.response)

    return self.config

# %% ../../nbs/classes/50_DomoAccount.ipynb 22
@patch_to(DomoAccount, cls_method=True)
async def get_from_id(
    cls,
    auth: dmda.DomoAuth,
    account_id: int,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_prn: bool = False
):
    """retrieves account metadata and attempts to retrieve config"""

    res = await account_routes.get_account_from_id(
        auth=auth, account_id=account_id, session=session, debug_api=debug_api
    )

    if return_raw:
        return res

    if not res.is_success:
        return None

    obj = res.response
    acc = cls._from_json(obj, auth)

    try:
        await acc._get_config(session=session, debug_api=debug_api, debug_prn = debug_prn)

    except DomoAccount_DataProviderType_ConfigNotDefined as e:
        print(e)
    
    except Exception as e:
        print(e)

    finally:
        return acc

# %% ../../nbs/classes/50_DomoAccount.ipynb 26
@patch_to(DomoAccount)
async def update_config(
    self: DomoAccount,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    config: DomoAccount_Config = None,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):

    auth = auth or self.auth

    config = config or self.config

    res = await account_routes.update_account_config(
        auth=auth,
        account_id=self.id,
        data_provider_type=self.data_provider_type,
        config_body=config.to_json(),
        debug_api=debug_api,
        session=session,
    )

    if return_raw:
        return res

    await self._get_config(session=session, debug_api=debug_api)

    return self


# %% ../../nbs/classes/50_DomoAccount.ipynb 29
class DomoAccount_UpdateName_Error(de.DomoError):
    def __init__(
        self,
        domo_instance,
        status,
        message,
        entity_id,
        function_name="update_name",
    ):

        super().__init__(
            function_name=function_name,
            entity_id=entity_id,
            domo_instance=domo_instance,
            status=status,
            message=message,
        )


@patch_to(DomoAccount)
async def update_name(
    self: DomoAccount,
    account_name: str = None,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):

    auth = auth or self.auth

    # print(auth, self.id, self.data_provider_type, self.config.to_json())

    res = await account_routes.update_account_name(
        auth=auth,
        account_id=self.id,
        account_name=account_name or self.name,
        debug_api=debug_api,
        session=session,
    )

    if return_raw:
        return res

    if not res.is_success:
        raise DomoAccount_UpdateName_Error(
            entity_id=self.id,
            domo_instance=auth.domo_instance,
            status=res.status,
            message=res.response,
        )

    self = await self.get_from_id(auth=auth, account_id=self.id)

    return self

# %% ../../nbs/classes/50_DomoAccount.ipynb 33
class DomoAccount_CreateAccount_Error(de.DomoError):
    def __init__(
        self,
        entity_id,
        domo_instance,
        status,
        message,
        function_name="create_account",
    ):

        super().__init__(
            function_name=function_name,
            entity_id=entity_id,
            domo_instance=domo_instance,
            status=status,
            message=message,
        )

# %% ../../nbs/classes/50_DomoAccount.ipynb 34
@patch_to(DomoAccount, cls_method=True)
def generate_create_body(cls, account_name, config):
    return {
        "displayName": account_name,
        "dataProviderType": config.data_provider_type,
        "name": config.data_provider_type,
        "configurations": config.to_json(),
    }


@patch_to(DomoAccount, cls_method=True)
async def create_account(
    cls: DomoAccount,
    account_name: str,
    config: DomoAccount_Config,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):

    body = cls.generate_create_body(account_name=account_name, config=config)

    res = await account_routes.create_account(
        auth=auth, config_body=body, debug_api=debug_api, session=session
    )

    if not res.is_success:
        raise DomoAccount_CreateAccount_Error(
            entity_id=account_name,
            domo_instance=auth.domo_instance,
            status=res.status,
            message=res.response,
        )

    return await cls.get_from_id(auth=auth, account_id=res.response.get("id"))

# %% ../../nbs/classes/50_DomoAccount.ipynb 35
class DomoAccount_DeleteAccount_Error(de.DomoError):
    def __init__(
        self,
        entity_id,
        domo_instance,
        status,
        message,
        function_name="delete_account",
    ):

        super().__init__(
            function_name=function_name,
            entity_id=entity_id,
            domo_instance=domo_instance,
            status=status,
            message=message,
        )


@patch_to(DomoAccount)
async def delete_account(
    self: DomoAccount,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):

    auth = auth or self.auth

    res = await account_routes.delete_account(
        auth=auth, account_id=self.id, debug_api=debug_api, session=session
    )

    if not res.is_success:

        raise DomoAccount_DeleteAccount_Error(
            entity_id=self.id,
            domo_instance=auth.domo_instance,
            status=res.status,
            message=res.response,
        )

    return True

# %% ../../nbs/classes/50_DomoAccount.ipynb 36
@patch_to(DomoAccount)
async def _is_group_ownership_beta(self, auth: dmda.DomoAuth):

    import domolibrary.classes.DomoBootstrap as dmbs

    domo_bsr = dmbs.DomoBootstrap(auth=auth or self.auth)
    
    domo_feature_ls = await domo_bsr.get_features()

    match_accounts_v2 = next(
        (domo_feature for domo_feature in domo_feature_ls if domo_feature.name == 'accounts-v2'), None)

    return True if match_accounts_v2 else False


@patch_to(DomoAccount)
async def share_account(
    self,
    user_id: int,
    auth: dmda.DomoAuth = None,
    is_v2: bool = None,
    access_level: ShareAccount = None,  # will default to Read
    debug_api: bool = False,
    debug_prn: bool = False,
    session: httpx.AsyncClient = None,
):
    auth = auth or self.auth

    if isinstance(auth, dmda.DomoFullAuth) and is_v2 is None:
        is_v2 = await self._is_group_ownership_beta(auth)

    if debug_prn:
        print(
            f"ℹ️ - {auth.domo_instance} - {'is' if is_v2 else 'is not'} v2_group_ownership")

    if is_v2 is None:
        raise Exception(
            """🛑 ERROR must explicitly pass a value for the `is_v2` boolean to share_accounts function.ABC
alternatively, use `dmda.DomoFullAuth` to automatically retrieve the correct setting.ABC
account sharing differs between v1 and v2 of the API.""")
        
        return None

    res = None

    if is_v2:
        share_payload = account_routes.generate_share_account_payload_v2(
            user_id=user_id, access_level=access_level or ShareAccount_V2_AccessLevel.CAN_VIEW
        )

        res = await account_routes.share_account_v2(
            auth=auth,
            account_id=self.id,
            share_payload=share_payload,
            debug_api=debug_api,
            session=session,
        )

    else:
        share_payload = account_routes.generate_share_account_payload_v1(
            user_id=user_id, access_level=access_level or ShareAccount_V1_AccessLevel.CAN_VIEW
        )

        res = await account_routes.share_account_v1(
            auth=auth,
            account_id=self.id,
            share_payload=share_payload,
            debug_api=debug_api,
            session=session,
        )

    if res.status == 500 and res.response == 'Internal Server Error':
        res.response = f'ℹ️ - {res.response + " | User may own account."}'

    if res.status == 200:
        res.response = f"shared {self.id} - {self.name} with {user_id}"

    return res

# %% ../../nbs/classes/50_DomoAccount.ipynb 39
@patch_to(DomoAccount)
async def share(
    self: DomoAccount,
    domo_user = None,
    domo_group = None,
    auth: dmda.DomoAuth = None,
    is_v2: bool = None,
    access_level: ShareAccount = None,  # will default to Read
    debug_api: bool = False,
    debug_prn: bool = False,
    session: httpx.AsyncClient = None,
):
    auth = auth or self.auth

    if isinstance(auth, dmda.DomoFullAuth) and is_v2 is None:
        is_v2 = await self._is_group_ownership_beta(auth)

    if debug_prn:
        print(
            f"ℹ️ - {auth.domo_instance} - {'is' if is_v2 else 'is not'} v2_group_ownership")

    if is_v2 is None:
        raise Exception(
            """🛑 ERROR must pass `is_v2` bool to share_accounts function IF NOT passing `dmda.DomoFullAuth`.
the group management v2 API has a different body.  
Alternatively pass a full auth object to auto check the bootstrap.
""")

    res = None

    if is_v2:
        share_payload = account_routes.generate_share_account_payload_v2(
            user_id=domo_user.id if domo_user else None, 
            group_id =  domo_group.id if domo_group else None,
            access_level=access_level or ShareAccount_V2_AccessLevel.CAN_VIEW
        )

        res = await account_routes.share_account_v2(
            auth=auth,
            account_id=self.id,
            share_payload=share_payload,
            debug_api=debug_api,
            session=session,
        )

    else:
        share_payload = account_routes.generate_share_account_payload_v1(
            user_id=domo_user.id if domo_user else None,
            group_id=domo_group.id if domo_group else None,
            
            access_level=access_level or ShareAccount_V1_AccessLevel.CAN_VIEW
        )

        res = await account_routes.share_account_v1(
            auth=auth,
            account_id=self.id,
            share_payload=share_payload,
            debug_api=debug_api,
            session=session,
        )

    if res.status == 500 and res.response == 'Internal Server Error':
        res.response = f'ℹ️ - {res.response + "| User may own account."}'

    if res.status == 200:
        domo_entity = domo_user or domo_group
        res.response = f"shared {self.id} - {self.name} with {domo_entity.id}"

    return res

# %% ../../nbs/classes/50_DomoAccount.ipynb 42
@dataclass
class DomoAccounts:
    auth: dmda.DomoAuth

# %% ../../nbs/classes/50_DomoAccount.ipynb 43
@patch_to(DomoAccounts, cls_method=True)
async def _get_accounts_accountsapi(
    cls: DomoAccounts,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):

    res = await account_routes.get_accounts(
        auth=auth, debug_api=debug_api, session=session
    )

    if return_raw or len(res.response) == 0:
        return res

    return await asyncio.gather(
        *[DomoAccount.get_from_id(
            account_id=json_obj.get("id"),
            debug_api=debug_api,
            session=session,
            auth=auth
        ) for json_obj in res.response])


@patch_to(DomoAccounts, cls_method=True)
async def _get_accounts_queryapi(
    cls: DomoAccounts,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    additional_filters_ls=None,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):

    """v2 api for works with group_account_v2 beta"""

    import domolibrary.routes.datacenter as datacenter_routes

    res = await datacenter_routes.search_datacenter(
        auth=auth,
        entity_type=datacenter_routes.Datacenter_Enum.ACCOUNT.value,
        additional_filters_ls=additional_filters_ls,
        session=session,
        debug_api=debug_api,
    )

    if return_raw or len(res.response) == 0:
        return res

    return [DomoAccount._from_json(account_obj, auth=auth) for account_obj in res.response]


@patch_to(DomoAccounts, cls_method=True)
async def get_accounts(
    cls: DomoAccounts,
    auth: dmda.DomoAuth,
    additional_filters_ls=None,  # datacenter_routes.generate_search_datacenter_filter
    # account string to search for, must be an exact match in spelling.  case insensitive
    # v2 will use the queryAPI as it returns more complete results than the accountsAPI
    is_v2: bool = None,
    account_name: str = None,
    account_type: AccountConfig = None,  # to retrieve a specific account type
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
    debug_prn: bool = False
):
    import domolibrary.classes.DomoBootstrap as bsr
    import domolibrary.routes.datacenter as datacenter_routes

    if isinstance(auth, dmda.DomoFullAuth) and is_v2 is None:
        instance_bsr = bsr.DomoBootstrap(auth=auth)

        is_v2 = await instance_bsr.is_group_ownership_beta(auth)

        if debug_prn:
            print(
                f"{auth.domo_instance} {'is' if is_v2 else 'is not'} using the v2 beta")

    if is_v2:
        try:
            domo_accounts = await cls._get_accounts_queryapi(
                auth=auth,
                debug_api=debug_api,
                additional_filters_ls=additional_filters_ls,
                session=session,
            )
        except datacenter_routes.SearchDatacenter_NoResultsFound as e:
            domo_accounts = []
    else:
        domo_accounts = await cls._get_accounts_accountsapi(
            auth=auth, debug_api=debug_api,
            session=session)

    if return_raw or len(domo_accounts) == 0:
        return domo_accounts

    if not account_name and not account_type:
        return domo_accounts

    if account_name and isinstance(account_name, str):
        domo_accounts = [
            domo_account
            for domo_account in domo_accounts if domo_account.name.lower() == account_name.lower()]

    if account_type:
        domo_accounts = [
            domo_account
            for domo_account in domo_accounts
            if domo_account.data_provider_type == account_type.value.data_provider_type
        ]

    return domo_accounts




import aiohttp

from .get_data import get_data
from ..DomoAuth import DomoFullAuth
from ...utils.ResponseGetData import ResponseGetData


async def get_roles(full_auth: DomoFullAuth,
                    debug: bool = False,
                    session: aiohttp.ClientSession = None,
                    ) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/authorization/v1/roles'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='GET',
        debug=debug,
        session=session
    )
    return res


async def create_role(full_auth: DomoFullAuth,
                      name: str,
                      description: str,
                      debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/authorization/v1/roles'

    body = {
        'name': name,
        'description': description
    }

    if debug:
        print(url)
        print({"body": body})

    res = await get_data(
        auth=full_auth,
        url=url,
        method='POST',
        debug=debug,
        body=body
    )

    return res


async def set_default_role(full_auth: DomoFullAuth,
                           role_id: str,
                           debug: bool = False,
                           session: aiohttp.ClientSession = None) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/content/v1/customer-states/user.roleid.default'

    body = {
        'name': 'user.roleid.default',
        'value': role_id
    }

    if debug:
        print(url)
        print({"body": body})

    res = await get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        debug=debug,
        body=body,
        session=session
    )
    return res


async def get_role_grants(full_auth: DomoFullAuth,
                          role_id: str,
                          debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/authorities'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='GET',
        debug=debug,
    )
    return res


async def update_role_grants(full_auth: DomoFullAuth,
                             role_id: str,
                             role_grant_list: list[str],
                             debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/authorities'

    if debug:
        print(url)

    res = await get_data(auth=full_auth,
                         url=url,
                         method='PUT',
                         debug=debug, body=role_grant_list)
    return res


async def role_membership_add_user(full_auth: DomoFullAuth,
                                   role_id: str,
                                   user_list: list[str],
                                   debug: bool = False) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/users'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        debug=debug,
        body=user_list
    )
    return res


async def get_role_membership(full_auth: DomoFullAuth,
                              role_id: str,
                              debug: bool = False, session: aiohttp.ClientSession = None) -> ResponseGetData:
    url = f'https://{full_auth.domo_instance}.domo.com/api/authorization/v1/roles/{role_id}/users'

    if debug:
        print(url)

    res = await get_data(
        auth=full_auth,
        url=url,
        method='GET',
        debug=debug,
        session=session
    )
    return res



from dataclasses import dataclass

import aiohttp

from . import DomoAuth as dmda
from .routes import get_data


async def _set_dataset_tags(full_auth: dmda.DomoFullAuth,
                            tag_ls: [str],
                            dataset_id: str,
                            debug: bool = False,
                            session: aiohttp.ClientSession = None,
                            log_results: bool = False
                            ):
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/ui/v3/datasources/{dataset_id}/tags"

    if debug:
        print(url)

    return await get_data(
        auth=full_auth,
        url=url,
        method='POST',
        log_results=log_results,
        debug=debug,
        body=tag_ls,
        session=session
    )


@dataclass
class Dataset_Tags:
    dataset: any
    tag_ls: [str] = None

    async def set(self,
                  tag_ls: [str],
                  full_auth: dmda.DomoFullAuth = None,
                  debug: bool = False, log_results: bool = False,
                  session: aiohttp.ClientSession = None):

        full_auth = full_auth or self.dataset.full_auth

        res = await _set_dataset_tags(full_auth=full_auth,
                                      tag_ls=list(set(tag_ls)),
                                      dataset_id=self.dataset.id,
                                      debug=debug,
                                      session=session,
                                      log_results=log_results)

        if res.status != 200:
            print('invalid response')
            return None

        new_ds = await self.dataset.get_from_id(id=self.dataset.id,
                                                full_auth=self.dataset.full_auth)

        self.tag_ls = new_ds.tags.tag_ls

        return self.tag_ls

    async def add(self,
                  add_tag_ls: [str],
                  full_auth: dmda.DomoFullAuth = None,
                  debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None
                  ):

        full_auth = full_auth or self.dataset.full_auth

        existing_tag_ls = self.tag_ls or []
        add_tag_ls += existing_tag_ls

        return await self.set(full_auth=full_auth,
                              tag_ls=list(set(add_tag_ls)),
                              debug=debug, session=session, log_results=log_results)

    async def remove(self,
                     remove_tag_ls: [str],
                     full_auth: dmda.DomoFullAuth = None,
                     debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None
                     ):
        full_auth = full_auth or self.dataset.full_auth

        existing_tag_ls = self.tag_ls or []
        existing_tag_ls = [
            ex for ex in existing_tag_ls if ex not in remove_tag_ls]

        return await self.set(full_auth=full_auth,
                              tag_ls=list(set(existing_tag_ls)),
                              debug=debug, session=session, log_results=log_results)



import aiohttp

from ...DomoClasses import DomoAuth as dmda
from . import get_data as gd


async def get_accounts(full_auth: dmda.DomoFullAuth,
                       debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):

    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/accounts"

    if debug:
        print(url)

    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='GET',
        log_results=log_results,
        debug=debug,
        session=session
    )


async def get_account_config(full_auth: dmda.DomoFullAuth,
                             account_id: int,
                             data_provider_type: bool = False,
                             debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):

    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/providers/{data_provider_type}/account/{account_id}?unmask=true"

    if debug:
        print(url)

    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='GET',
        log_results=log_results,
        debug=debug,
        session=session
    )


async def update_account_config(full_auth: dmda.DomoFullAuth,
                                account_id: int,
                                config_body: dict,
                                data_provider_type: str,
                                debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):

    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/providers/{data_provider_type}/account/{account_id}"

    if debug:
        print(url)

    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        body=config_body,
        log_results=log_results,
        debug=debug,
        session=session
    )

async def update_account_name(full_auth: dmda.DomoFullAuth,
                              account_id: int,
                              account_name: str,
                              debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):

    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}/name"
    
    if debug:
        print(url)

    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        body=account_name,
        log_results=log_results,
        content_type = "text/plain",
        debug=debug,
        session=session
    )





async def get_account_config(full_auth: dmda.DomoFullAuth,
                             account_id: int,
                             data_provider_type: bool = False,
                             debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):

    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/providers/{data_provider_type}/account/{account_id}?unmask=true"

    if debug:
        print(url)

    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='GET',
        log_results=log_results,
        debug=debug,
        session=session
    )


async def get_account_from_id(full_auth: dmda.DomoFullAuth, account_id: int,
                              debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}?unmask=true"

    if debug:
        print(url)

    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='GET',
        log_results=log_results,
        debug=debug,
        session=session
    )


async def create_account(full_auth:dmda.DomoFullAuth, config_body:dict,
                         debug: bool = False, log_results: bool = False, session: aiohttp.ClientSession = None):
    
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/accounts"

    if debug:
        print(url)

    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='POST',
        body = config_body,
        log_results=log_results,
        debug=debug,
        session=session
    )

async def delete_account(full_auth:dmda.DomoFullAuth,
                         account_id: str,
                         debug: bool = False, 
                         log_results: bool = False, session: aiohttp.ClientSession = None):
    
    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}"

    if debug:
        print(url)

    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='DELETE',
        log_results=log_results,
        debug=debug,
        session=session
    )

def generate_share_account_payload_v1( user_id: int ):
    return {"type":"USER","id":user_id,"permissions":['READ']}

def generate_share_account_payload_v2( user_id: int,
                                   access_level: str = 'CAN_VIEW' # CAN_VIEW, CAN_EDIT, CAN_SHARE
                                  ):
    return {"type":"USER","id":user_id,"accessLevel":access_level }



async def share_account_v2(full_auth : dmda.DomoFullAuth,
                        account_id : str,
                        share_payload: dict,
                        debug: bool = False,
                        log_results :bool = False,
                        session : aiohttp.ClientSession = None
                       ):

    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v2/accounts/share/{account_id}"
    
    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        body = share_payload,
        log_results=log_results,
        debug=debug,
        session=session
    )

async def share_account_v1(full_auth : dmda.DomoFullAuth,
                        account_id : str,
                        share_payload: dict,
                        debug: bool = False,
                        log_results :bool = False,
                        session : aiohttp.ClientSession = None
                       ):

    url = f"https://{full_auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}/share"
    
    return await gd.get_data(
        auth=full_auth,
        url=url,
        method='PUT',
        body = share_payload,
        log_results=log_results,
        debug=debug,
        session=session
    )






from dataclasses import dataclass, field

from . import Base as b


@dataclass
class ResponseGetData(b.Base):
    status: int
    response: any
    is_success: bool
    auth: dict = field(default=None)

    def __post_init__(self):
        super().__init__()
